# ------------------------------------------------------------------------------
#  es7s/commons
#  (c) 2022-2023 A. Shavykin <0.delameter@gmail.com>
# ------------------------------------------------------------------------------
from abc import ABCMeta, abstractmethod
from functools import cached_property


class Spinner(metaclass=ABCMeta):
    def __init__(self):
        self._frame = 0
        self._max_frame = len(self._frames)

    @property
    @abstractmethod
    def _frames(self) -> list[str]:
        raise NotImplementedError

    def render(self, string: str):
        frame = self._frames[self._frame]
        self._frame += 1
        if self._frame >= self._max_frame:
            self._frame = 0
        return f"\r{frame} {string}"


class SpinnerBrailleSquare(Spinner):
    @cached_property
    def _frames(self) -> list[str]:
        # fmt: off
        return [
            "⡀ ", "⠄ ", "⠂ ", "⠁ ",
            "⠈ ", " ⠁", " ⠈", " ⠐",
            " ⠠", " ⢀", " ⡀", "⢀ ",
        ]
        # fmt: on


class SpinnerBrailleSquareCenter(Spinner):
    @cached_property
    def _frames(self) -> list[str]:
        # fmt: off
        return [
            "⡰⠆", "⠴⠆", "⠲⠆", "⠱⠆",
            "⠸⠆", "⠰⠇", "⠰⠎", "⠰⠖",
            "⠰⠦", "⠰⢆", "⠰⡆", "⢰⠆",
        ]
        # fmt: on


class SpinnerBrailleSquareFill(Spinner):
    @cached_property
    def _frames(self) -> list[str]:
        # fmt: off
        return [
            "⢿⣿", "⣻⣿", "⣽⣿", "⣾⣿",
            "⣷⣿", "⣿⣾", "⣿⣷", "⣿⣯",
            "⣿⣟", "⣿⡿", "⣿⢿", "⡿⣿",
        ]
        # fmt: on



# ------------------------------------------------------------------------------
#  es7s/commons
#  (c) 2023 A. Shavykin <0.delameter@gmail.com>
# ------------------------------------------------------------------------------
from __future__ import annotations

import typing as t
from functools import update_wrapper
from logging import getLogger, DEBUG
from typing import cast, Optional, Union, overload

import pytermor as pt

from .common import nowf

_F = t.TypeVar("_F", bound=t.Callable[..., t.Any])
_MFT = t.TypeVar("_MFT", bound=t.Callable[[str, t.Any, ...], Optional[Union[str, t.Iterable[str]]]])


@overload
def measure(__origin: _F) -> _F:
    ...

@overload
def measure(*, fmter: _MFT = None, level=DEBUG) -> t.Callable[[_F], _F]:
    ...

def measure(__origin: _F = None, *, fmter: _MFT = None, level=DEBUG) -> _F | t.Callable[[_F], _F]:
    def _default_formatter(delta_s: str, *_, **__) -> t.Iterable[str]:
        yield f"Done in {delta_s}"

    def decorator(origin: t.Callable[..., t.Any]):
        def wrapper(*args, **kwargs):
            logger = getLogger(__package__)
            before_s = nowf()
            result = origin(*args, **kwargs)
            delta_s = nowf() - before_s

            try:
                fmt_fn: _MFT = fmter or _default_formatter
                if msg := fmt_fn(_format_sec(delta_s), result, *args, **kwargs):
                    for m in msg if pt.isiterable(msg) else [msg]:
                        logger.log(level=level, msg=m)
            except Exception as e:  # pragma: no cover
                logger.exception(e)
            return result

        return update_wrapper(cast(_F, wrapper), origin)

    if __origin is not None:
        return decorator(__origin)
    else:
        return decorator


def _trace_render(origin: _F) -> _F:
    def measure_format_in_out(delta_s: str, out: t.Any, *args, **_) -> t.Iterable[str]:
        actor, inp, extra = args
        no_changes = out == inp
        inp_start = pt.cut(inp, 40)
        out_start = pt.cut(out, 40)
        if no_changes:
            yield f"○ {pt.get_qname(actor)} noop in {delta_s} ({len(inp)}): {inp_start!r}"
        else:
            inplen, outlen = (str(len(s or "")) for s in [inp, out])
            maxlen = max(len(inplen), len(outlen))
            yield f"╭ {actor!r} applying {extra!r}"
            yield f"│ IN  ({inplen:>{maxlen}s}): {inp_start!r}"
            yield f"│ OUT ({outlen:>{maxlen}s}): {out_start!r}"
            yield f"╰ {delta_s}"

    @measure(fmter=measure_format_in_out)
    def new_func(*args, **kwargs):
        return origin(*args, **kwargs)

    return update_wrapper(cast(_F, new_func), origin)


def _format_sec(val: float) -> str:  # pragma: no cover
    if val >= 2:
        return f"{val:.1f}s"
    if val >= 2e-3:
        return f"{val*1e3:.0f}ms"
    if val >= 2e-6:
        return f"{val*1e6:.0f}µs"
    if val >= 1e-9:
        return f"{val*1e9:.0f}ns"
    return "<1ns"



# ------------------------------------------------------------------------------
#  es7s/commons
#  (c) 2023 A. Shavykin <0.delameter@gmail.com>
# ------------------------------------------------------------------------------
from __future__ import annotations

import io
import os
import typing as t
from collections import deque
from functools import partial
from logging import getLogger
from math import ceil
from pathlib import Path

import pytermor as pt
from pytermor import filtern, FT


def joincoal(*arg: any, sep="") -> str:
    return sep.join(map(str, filtern(arg)))


isempty = lambda v: bool(v) and len(str(v).strip())


filtere = partial(filter, isempty)
""" Shortcut for filtering out falsy AND empty values from sequences """


def filterev(mapping: dict) -> dict:
    """Shortcut for filtering out falsy AND empty values from mappings"""
    return dict(filter(lambda kv: isempty(kv[1]), mapping.items()))


class DisposableComposite(pt.Composite):
    pass
    # def render(self, *args) -> str:
    #    return super().render(*args).replace(' ', '_')


class AdaptiveFragment(pt.Fragment):
    COLLAPSE_CHAR = " "
    COLLAPSE_MAX_LVL = 4
    COLLAPSE_FNS = {
        1: lambda s: s.removeprefix,
        2: lambda s: s.removesuffix,
        3: lambda s: s.lstrip,
        4: lambda s: s.rstrip,
    }

    def __init__(self, min_len: int, string: str = "", fmt: FT = None):
        super().__init__(string, fmt)
        self._min_len = min_len
        self._collapse_lvl = 0

    def collapse(self, lvl: int = 0):
        self._collapse_lvl = lvl
        self._string = self.collapsed(lvl)

    def collapsed(self, lvl: int = 0) -> str:
        collapse_fn = self.COLLAPSE_FNS.get(lvl, lambda s: s.strip)
        return collapse_fn(self._string)(self.COLLAPSE_CHAR)

    def delta_collapse(self, lvl: int = 0) -> int:
        return len(self._string) - len(self.collapsed(lvl))

    @property
    def collapse_lvl(self) -> int:
        return self._collapse_lvl

    def shrink(self):
        self._string = self.shrinked()

    def shrinked(self) -> str:
        return self._string[: self._min_len]

    def delta_shrink(self):
        return len(self._string) - len(self.shrinked())


_AF = AdaptiveFragment
_DC = DisposableComposite


class CompositeCompressor(pt.Composite):
    def __init__(self, *parts: pt.RT):
        super().__init__(*parts)

    def append(self, part: pt.RT):
        self._parts.append(part)

    def extend(self, parts: t.Iterable[pt.RT]):
        self._parts.extend(parts)

    def compress(self, max_len: int):
        """
        5 levels of elements compression, from almost transparent to monstrously barbaric:

        * purge       delete special blank "disposable" pseudo-fragments that act as a ballast;
        * collapse    remove whitespace characters marked as removable in adaptive fragments;
        * shrink      decrease the length of adaptive fragments to minimum recommended values,
                      where the labels can still be comprehended (usually 3-4 letters);
        * chop        ignore minimum recommended value, forcefully cut out last characters of
                      the fragments that can be modified externally; try to distribute the
                      resections evenly between all fragments to make them distinguishable from
                      each other as long as possible;
        * eviscerate  throw away the rightmost fragments entirely, which is the only remaining
                      method if the external modification is not supported by element interfaces;
                      this allows to keep at least some parts of some of fragments.
        """
        if max_len == 0:
            self._parts.clear()
            return
        req_delta = len(self) - max_len
        if req_delta <= 0:
            return
        if len(self._parts[0]) > max_len:
            self._parts = deque[pt.IRenderable]([self._parts.popleft()])

        disposables = []
        adaptives = []
        fragments = []

        for part in self._parts:
            if isinstance(part, _DC):
                disposables.append(part)
            elif isinstance(part, _AF):
                adaptives.append(part)
                fragments.append(part)
            elif isinstance(part, pt.Fragment):
                fragments.append(part)

        cur_pack_lvl = 1

        while (req_delta := len(self) - max_len) > 0:
            if max_purge := sum(map(len, disposables)):
                self._purge(disposables, req_delta, max_purge)
                self._debug_compress_level("I")
                disposables.clear()
                continue

            if cur_pack_lvl < _AF.COLLAPSE_MAX_LVL and sum(
                map(lambda af: af.delta_collapse(cur_pack_lvl), adaptives)
            ):
                self._collapse(adaptives, cur_pack_lvl)
                self._debug_compress_level(f"II ({cur_pack_lvl})")
                cur_pack_lvl += 1
                continue

            if max_shrink := sum(map(_AF.delta_shrink, adaptives)):
                self._shrink(adaptives, req_delta - max_shrink)
                self._debug_compress_level("III")
                adaptives.clear()
                continue

            if max_chop := sum(map(len, fragments)):
                self._chop(fragments, req_delta, max_chop)
                self._debug_compress_level("IV")
                fragments.clear()
                continue

            self._eviscerate(max_len)
            self._debug_compress_level("V")
            break  # от греха

    def _debug_compress_level(self, level: str):
        getLogger(__package__).debug(f"Level {level} compression applied: length {len(self)}")

    def _purge(self, disposables: list[_DC], req_delta: int, max_purge: int):
        if req_delta - max_purge > 0:
            for d in disposables:
                self._parts.remove(d)
            return
        for d in sorted(disposables, key=lambda D: -len(D)):
            if req_delta <= 0:
                break
            req_delta -= len(d)
            self._parts.remove(d)

    def _collapse(self, adaptives: list[_AF], lvl: int):
        [a.collapse(lvl) for a in adaptives]

    def _shrink(self, adaptives: list[_AF], req_remain: int):
        [a.shrink() for a in adaptives]

    def _chop(self, fragments: list[pt.Fragment], req_delta: int, delta_chop: int):
        if req_delta - delta_chop > 0:
            for f in fragments:
                self._parts.remove(f)
            return
        chop_ratio = req_delta / len(fragments)
        for f in fragments:
            if req_delta <= 0:
                break
            req_delta -= (chop := ceil(len(f) * chop_ratio))
            if chop >= len(f):
                self._parts.remove(f)
            else:
                f._string = f._string[:-chop]

    def _eviscerate(self, max_len: int):
        while len(self) > max_len and len(self._parts):
            self._parts.pop()


def format_attrs(*o: object, keep_classname=True, level=0, flat=False) -> str:
    def _to_str(a) -> str:
        if (s := str(a)).startswith(cn := a.__class__.__name__):
            if keep_classname:
                return s
            return s.removeprefix(cn)
        return f"'{s}'" if s.count(" ") else s

    def _wrap(s):
        if flat:
            return s
        return f"({s})"

    if len(o) == 1:
        o = o[0]
    if isinstance(o, str):
        return o
    elif isinstance(o, t.Mapping):
        return _wrap(" ".join(f"{_to_str(k)}={format_attrs(v, flat=flat)}" for k, v in o.items()))
    elif issubclass(type(o), io.IOBase):
        return f"{pt.get_qname(o)}['{getattr(o, 'name', '?')}', {getattr(o, 'mode', '?')}]"
    elif isinstance(o, t.Iterable):
        return _wrap(" ".join(format_attrs(v, level=level + 1, flat=flat) for v in o))
    return _to_str(o)


def format_path(path: str | Path, *, color=False, repr=True) -> str | pt.RT:
    fmt = "{!r}" if repr else "{:s}"
    apath = os.path.abspath(str(path))
    result = pt.Text(fmt.format(apath), pt.cv.BLUE)
    if (rpath := os.path.realpath(apath)) != apath:
        result += " -> " + pt.Fragment(fmt.format(rpath), pt.cv.YELLOW)
    if not color:
        result = result.raw()
    return result



# ------------------------------------------------------------------------------
#  es7s/commons
#  (c) 2023 A. Shavykin <0.delameter@gmail.com>
# ------------------------------------------------------------------------------
import sys
import typing as t
from logging import getLogger

import pytermor as pt


class InputMode(str, pt.ExtendedEnum):  # @TODO str enums will be available in python 3.11
    DEFAULT: str = 'default'
    DISABLED: str = 'disabled'
    DISCRETE: str = 'discrete'


class TerminalState:
    def __init__(self, io: t.IO = sys.stdout):
        self._io = io
        self._restore_callbacks: list[t.Callable[[], None]] = []
        self._terminal_orig_settings: list | None = None

    def enable_alt_screen_buffer(self):
        self._add_restore_callback(self.disable_alt_screen_buffer)
        self._add_restore_callback(self._restore_cursor_position)
        self._send_sequence(pt.make_save_cursor_position)
        self._send_sequence(pt.make_enable_alt_screen_buffer)

    def disable_alt_screen_buffer(self):
        self._remove_restore_callback(self.disable_alt_screen_buffer)
        self._send_sequence(pt.make_disable_alt_screen_buffer)
        getLogger(__package__).debug("TSC: DISABLED ALT SCREEN BUFFER: stderr logging should work again")

    def hide_cursor(self):
        self._add_restore_callback(self.show_cursor)
        self._send_sequence(pt.make_hide_cursor)

    def show_cursor(self):
        self._remove_restore_callback(self.show_cursor)
        self._send_sequence(pt.make_show_cursor)

    def setup_input(self, input_mode: InputMode):
        match input_mode:
            case InputMode.DEFAULT:
                self.restore_input()
            case InputMode.DISABLED:
                self.disable_input()
            case InputMode.DISCRETE:
                self.discrete_input()
            case _:
                raise NotImplementedError(f"Unknown input mode: {input_mode}")

    def disable_input(self):
        getLogger(__package__).debug(f"TSC: Putting tty into cbreak mode: {sys.stdin}")
        if not self._is_a_tty():
            return
        if self._terminal_orig_settings:
            getLogger(__package__).warning(f"TSC: Altering tty attributes skipped: already altered")
            return

        import tty
        import termios

        self._terminal_orig_settings = termios.tcgetattr(sys.stdin)
        self._add_restore_callback(self.restore_input)
        try:
            tty.setcbreak(sys.stdin)
        except (termios.error, TypeError) as e:
            getLogger(__package__).error("Saving tty state failed", exc_info=e)

    def discrete_input(self):
        """
        Use with: sys.stdin.read(1)
        Origin: readchar/_posix_read.py
        """
        getLogger(__package__).debug(f"TSC: Enabling tty discrete input: {sys.stdin}")
        if not self._is_a_tty():
            return
        if self._terminal_orig_settings:
            getLogger(__package__).warning(f"TSC: Altering tty attributes skipped: already altered")
            return

        import termios

        fd = sys.stdin.fileno()
        term = termios.tcgetattr(fd)
        self._terminal_orig_settings = termios.tcgetattr(fd)
        self._add_restore_callback(self.restore_input)
        try:
            # originally:
            # term[3] &= ~(termios.ICANON | termios.ECHO)
            term[3] &= ~(termios.ICANON | termios.ECHO | termios.IGNBRK | termios.BRKINT)
            # we don't need to ignore Ctrl+C though, because we handle it on the top level in init_cli()
            termios.tcsetattr(fd, termios.TCSAFLUSH, term)
        except (termios.error, TypeError) as e:
            getLogger(__package__).error("Saving tty state failed", exc_info=e)

    def restore_input(self):
        if not self._is_a_tty():
            getLogger(__package__).warning(f"TSC: Restoring tty attributes skipped: not a tty")
            return
        if not self._terminal_orig_settings:
            getLogger(__package__).warning(f"TSC: Restoring tty attributes skipped: nothing to restore")
            return

        getLogger(__package__).debug(f"TSC: Restoring tty attributes: {self._terminal_orig_settings}")
        self._remove_restore_callback(self.restore_input)

        import termios

        try:
            termios.tcsetattr(sys.stdin, termios.TCSADRAIN, self._terminal_orig_settings)
        except TypeError as e:
            getLogger(__package__).error("Restoring tty state failed", exc_info=e)
        finally:
            self._terminal_orig_settings = None

    def _restore_cursor_position(self):
        self._remove_restore_callback(self._restore_cursor_position)
        self._send_sequence(pt.make_restore_cursor_position)

    def restore_state(self):
        while self._restore_callbacks:
            getLogger(__package__).debug(f"TSC: Restoring state ({len(self._restore_callbacks)})")
            try:
                self._restore_callbacks.pop()()
            except:  # noqa
                pass

    def _add_restore_callback(self, fn: t.Callable[[], None]) -> None:
        getLogger(__package__).debug(f"TSC: Registering callback: '{fn.__name__}'")
        self._restore_callbacks.append(fn)

    def _remove_restore_callback(self, fn: t.Callable[[], None]) -> None:
        if fn not in self._restore_callbacks:
            return
        getLogger(__package__).debug(f"TSC: Unregistering callback: '{fn.__name__}'")
        self._restore_callbacks.remove(fn)

    def _send_sequence(self, fn: t.Callable[[], pt.ISequence]) -> None:
        if not self._is_a_tty():
            getLogger(__package__).debug(f"TSC: Skipping control sequence: not a tty")
            return

        sequence = fn()
        getLogger(__package__).debug(f"TSC: Sending control sequence: '{fn.__name__}' -> {sequence!r}")

        if sequence == pt.make_enable_alt_screen_buffer():
            getLogger(__package__).debug(
                "TSC: ENABLED ALT SCREEN BUFFER: ALL STDERR LOG MESSAGES ARE NOW _IGNORED_ BY THE "
                "TERMINAL (syslog should still work unless switched off explicitly)"
            )
        self._echo(sequence)

    def _echo(self, sequence: pt.ISequence):
        self._io.write(sequence.assemble())

    def _is_a_tty(self):
        return self._io.isatty()



# ------------------------------------------------------------------------------
#  es7s/commons
#  (c) 2023 A. Shavykin <0.delameter@gmail.com>
# ------------------------------------------------------------------------------

import datetime
from dataclasses import dataclass
import pytermor as pt

WWO_CODE = {
    "113": "Sunny",
    "116": "PartlyCloudy",
    "119": "Cloudy",
    "122": "VeryCloudy",
    "143": "Fog",
    "176": "LightShowers",
    "179": "LightSleetShowers",
    "182": "LightSleet",
    "185": "LightSleet",
    "200": "ThunderyShowers",
    "227": "LightSnow",
    "230": "HeavySnow",
    "248": "Fog",
    "260": "Fog",
    "263": "LightShowers",
    "266": "LightRain",
    "281": "LightSleet",
    "284": "LightSleet",
    "293": "LightRain",
    "296": "LightRain",
    "299": "HeavyShowers",
    "302": "HeavyRain",
    "305": "HeavyShowers",
    "308": "HeavyRain",
    "311": "LightSleet",
    "314": "LightSleet",
    "317": "LightSleet",
    "320": "LightSnow",
    "323": "LightSnowShowers",
    "326": "LightSnowShowers",
    "329": "HeavySnow",
    "332": "HeavySnow",
    "335": "HeavySnowShowers",
    "338": "HeavySnow",
    "350": "LightSleet",
    "353": "LightShowers",
    "356": "HeavyShowers",
    "359": "HeavyRain",
    "362": "LightSleetShowers",
    "365": "LightSleetShowers",
    "368": "LightSnowShowers",
    "371": "HeavySnowShowers",
    "374": "LightSleetShowers",
    "377": "LightSleet",
    "386": "ThunderyShowers",
    "389": "ThunderyHeavyRain",
    "392": "ThunderySnowShowers",
    "395": "HeavySnowShowers",
}


@dataclass
class DynamicIcon:
    day_icon: str
    night_icon: str
    extra_icon: str | None = None

    def select(self, override_hour: int = None) -> str:  # @fixme use sun calc
        eff_hour = override_hour or datetime.datetime.now().hour
        if eff_hour == 0:
            return self.extra_icon or self.night_icon
        if eff_hour >= 22 or eff_hour <= 6:
            return self.night_icon
        return self.day_icon

    def get_raw(self) -> tuple[str, ...]:
        return tuple(filter(None, (self.day_icon, self.night_icon, self.extra_icon)))


class WeatherIconSet:
    NO_COLOR_SET_IDS = [0]
    MAX_SET_ID = 0

    def __init__(self, color_code: int, *icons: str | tuple[str, ...], wwo_codes: list[str]):
        self._style: pt.Style = pt.Style(fg=pt.Color256.get_by_code(color_code))
        self._icons: list[str | DynamicIcon] = [
            s if isinstance(s, str) else DynamicIcon(*s) for s in icons
        ]
        self._wwo_codes: list[str] = wwo_codes
        WeatherIconSet.MAX_SET_ID = max(WeatherIconSet.MAX_SET_ID, len(self._icons) - 1)

    def get_icon(self, set_id: int, override_hour: int = None) -> tuple[str, str, pt.Style]:
        """
        :return: (icon, terminator, style)
        """
        if set_id < 0 or set_id > WeatherIconSet.MAX_SET_ID:
            raise IndexError(f"Set #{set_id} is undefined")

        icon = self._icons[set_id]
        if isinstance(icon, DynamicIcon):
            icon = icon.select(override_hour)

        if set_id in self.NO_COLOR_SET_IDS:
            return icon, WEATHER_ICON_TERMINATOR, pt.NOOP_STYLE
        return icon, WEATHER_ICON_TERMINATOR, self._style

    def get_raw(self, set_id: int) -> tuple[str]:
        if set_id < 0 or set_id >= WeatherIconSet.MAX_SET_ID:
            raise IndexError(f"Set #{set_id} is undefined")
        icon = self._icons[set_id]
        if isinstance(icon, DynamicIcon):
            return icon.get_raw()
        return (icon,)


# fmt: off
WEATHER_ICON_SETS: dict[str, WeatherIconSet] = {
    "✨": WeatherIconSet(248,	"✨️",	"",	"",	"",	("",	"",	""),	wwo_codes=["Unknown"]),
    "☀": WeatherIconSet(248,	"☀️",	"滛",	"滛",	"",	("",	"望",	""),	wwo_codes=["Sunny"]),
    "☁":  WeatherIconSet(248,	"☁️",	"摒",	"摒",	"",	("",	""),			wwo_codes=["Cloudy", "VeryCloudy"]),
    "⛅": WeatherIconSet(248,	"⛅️",	"杖",	"杖",	"",	("",	""), 			wwo_codes=["PartlyCloudy"]),
    "🌫": WeatherIconSet(248,	"🌫️",	"敖",	"敖",	"",	("",	""), 			wwo_codes=["Fog"]),
    "🌦": WeatherIconSet( 74,	"🌦️",	"",	"殺",	"",	("",	""), 			wwo_codes=["LightRain", "LightShowers"]),
    "🌧": WeatherIconSet( 74,	"🌧️",	"",	"殺",	"",	("",	""), 			wwo_codes=["HeavyRain", "HeavyShowers", "LightSleet", "LightSleetShowers"]),
    "⛈": WeatherIconSet(179,	"⛈️",	"",	"ﭼ",	"",	("",	""), 			wwo_codes=["ThunderyShowers", "ThunderySnowShowers"]),
    "🌩": WeatherIconSet(179,	"🌩️",	"",	"朗",	"",	("",	""),			wwo_codes=["ThunderyHeavyRain"]),
    "🌨": WeatherIconSet(153,	"🌨️",	"ﰕ",	"流",	"",	("",	""),			wwo_codes=["LightSnow", "LightSnowShowers"]),
    "❄": WeatherIconSet(153,	"❄️",	"",	"",	"",	("",	""),			wwo_codes=["HeavySnow", "HeavySnowShowers"]),
}

WEATHER_ICON_TERMINATOR = '\u200e'
# Unicode LRM to force disable right-to-left
# mode after some icons getting printed

WIND_DIRECTION = ["↓", "↙", "←", "↖", "↑", "↗", "→", "↘"]

WEATHER_SYMBOL_PLAIN = {
    "Unknown":				"?",
    "Cloudy":				"mm",
    "Fog":					"=",
    "HeavyRain":			"///",
    "HeavyShowers":			"//",
    "HeavySnow":			"**",
    "HeavySnowShowers":		"*/*",
    "LightRain":			"/",
    "LightShowers":			".",
    "LightSleet":			"x",
    "LightSleetShowers":	"x/",
    "LightSnow":			"*",
    "LightSnowShowers":		"*/",
    "PartlyCloudy":			"m",
    "Sunny":				"o",
    "ThunderyHeavyRain":	"/!/",
    "ThunderyShowers":		"!/",
    "ThunderySnowShowers":	"*!*",
    "VeryCloudy":			"mmm",
}
# fmt: on


def get_wicon(origin: str, set_id: int) -> tuple[str, str, pt.Style]:
    """
    :param origin: initial icon (wttr.in)
    :param set_id:
    :return: (updated-icon, string terminator, style)
    """
    if icon_set := WEATHER_ICON_SETS.get(origin, None):
        try:
            icon, term, style = icon_set.get_icon(set_id)
            return icon, term, style
        except IndexError:
            pass
    return origin, "", pt.NOOP_STYLE


def justify_wicon(icon: str, max_width: int, measure=False) -> tuple[str, int]:
    """
    :param icon:
    :param measure:
    :return: Tuple of justified icon and real icon width (if measure is *True*,
             they can differ).
    """
    if measure:
        width = pt.measure_char_width(icon)
    else:
        width = sum(max(0, pt.guess_char_width(c)) for c in icon)
    fill = "␣" if measure else " "
    return icon + fill * (max_width - width), width



# ------------------------------------------------------------------------------
#  es7s/commons
#  (c) 2023 A. Shavykin <0.delameter@gmail.com>
# ------------------------------------------------------------------------------
import typing as t
from math import floor

import pytermor as pt

from .strutil import to_superscript

FULL_BLOCK = "█"


class Scale(pt.Text):
    SCALE_LEN = 10

    def __init__(
        self,
        ratio: float,
        label_st: pt.FT,
        scale_st: pt.FT,
        length: int = SCALE_LEN,
        allow_partials: bool = True,
        full_block_char: str = FULL_BLOCK,
        start_char: str = None,
        require_not_empty: bool = False,
    ):
        self._ratio = ratio
        self._label_st = label_st
        self._scale_st = scale_st
        self._length = length
        self._allow_partials = allow_partials
        self._full_block_char = full_block_char
        self._start_char = start_char
        self._require_not_empty = require_not_empty

        self.label: str
        self.blocks: str
        super().__init__(*self._make())

    def _make(self) -> t.Iterable[pt.Fragment]:
        ratio_str = pt.format_auto_float(100 * self._ratio, 3)
        if ratio_str == "0.0":
            ratio_str = "e-2"
        if "e" in ratio_str:
            base, exp, power = ratio_str.partition("e")
            ratio_str = base + "10" + to_superscript(power)
        label_str = f"{ratio_str:>4s}% "
        self.label = pt.Fragment(" " + label_str, self._label_st)

        char_num: float = self._length * self._ratio
        full_block_num = floor(char_num)
        blocks_str = self._full_block_char * full_block_num
        if len(blocks_str) and self._start_char:
            blocks_str = self._start_char + blocks_str[1:]
        if self._allow_partials:
            blocks_str += get_partial_hblock(char_num - full_block_num)
        if not blocks_str and self._require_not_empty:
            blocks_str = "▏"
        self.blocks = pt.Fragment(blocks_str, self._scale_st)

        yield self.label
        yield self.blocks
        yield pt.Fragment(" " * (self._length - len(self.blocks)), self._scale_st)


def get_partial_hblock(val: float) -> str:  # @REFACTOR ME
    if val >= 7 / 8:
        return "▉"
    elif val >= 6 / 8:
        return "▊"
    elif val >= 5 / 8:
        return "▋"
    elif val >= 4 / 8:
        return "▌"
    elif val >= 3 / 8:
        return "▍"
    elif val >= 2 / 8:
        return "▎"
    elif val >= 1 / 8:
        return "▏"
    return ""
